/* valacodecontext.vala
 *
 * Copyright (C) 2006-2008  Jürg Billeter
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
 * Author:
 * 	Jürg Billeter <j@bitron.ch>
 */

#include <vala/valacodecontext.h>
#include <gee/arraylist.h>
#include <gee/list.h>
#include <gee/readonlycollection.h>
#include <glib/gstdio.h>
#include <config.h>
#include <vala/valamethod.h>
#include <vala/valasourcefile.h>
#include <vala/valanamespace.h>
#include <vala/valasourcereference.h>
#include <vala/valasourcefilecycle.h>
#include <vala/valasymbol.h>
#include <vala/valacodegenerator.h>
#include <vala/valacodevisitor.h>
#include <vala/valaattribute.h>




struct _ValaCodeContextPrivate {
	char* _library;
	gboolean _memory_management;
	gboolean _assert;
	gboolean _checking;
	gboolean _non_null;
	gboolean _non_null_experimental;
	gboolean _ccode_only;
	gboolean _compile_only;
	char* _output;
	char* _basedir;
	char* _directory;
	gboolean _debug;
	gint _optlevel;
	gboolean _thread;
	ValaMethod* _module_init_method;
	gboolean _save_temps;
	GeeList* source_files;
	GeeList* c_source_files;
	ValaNamespace* _root;
	GeeList* cycles;
	GeeList* packages;
	GeeList* defines;
	GeeList* dynamic_members;
	ValaCodeGenerator* _codegen;
};

#define VALA_CODE_CONTEXT_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), VALA_TYPE_CODE_CONTEXT, ValaCodeContextPrivate))
enum  {
	VALA_CODE_CONTEXT_DUMMY_PROPERTY,
	VALA_CODE_CONTEXT_LIBRARY,
	VALA_CODE_CONTEXT_MEMORY_MANAGEMENT,
	VALA_CODE_CONTEXT_ASSERT,
	VALA_CODE_CONTEXT_CHECKING,
	VALA_CODE_CONTEXT_NON_NULL,
	VALA_CODE_CONTEXT_NON_NULL_EXPERIMENTAL,
	VALA_CODE_CONTEXT_CCODE_ONLY,
	VALA_CODE_CONTEXT_COMPILE_ONLY,
	VALA_CODE_CONTEXT_OUTPUT,
	VALA_CODE_CONTEXT_BASEDIR,
	VALA_CODE_CONTEXT_DIRECTORY,
	VALA_CODE_CONTEXT_DEBUG,
	VALA_CODE_CONTEXT_OPTLEVEL,
	VALA_CODE_CONTEXT_THREAD,
	VALA_CODE_CONTEXT_MODULE_INIT_METHOD,
	VALA_CODE_CONTEXT_SAVE_TEMPS,
	VALA_CODE_CONTEXT_SAVE_CSOURCES,
	VALA_CODE_CONTEXT_SAVE_CHEADERS,
	VALA_CODE_CONTEXT_ROOT,
	VALA_CODE_CONTEXT_CODEGEN
};
static ValaSourceFile* vala_code_context_find_cycle_head (ValaCodeContext* self, ValaSourceFile* file);
static void vala_code_context_visit (ValaCodeContext* self, ValaSourceFile* file, GeeCollection* chain);
static GObject * vala_code_context_constructor (GType type, guint n_construct_properties, GObjectConstructParam * construct_properties);
static gpointer vala_code_context_parent_class = NULL;
static void vala_code_context_dispose (GObject * obj);



ValaCodeContext* vala_code_context_new (void) {
	ValaCodeContext * self;
	self = g_object_newv (VALA_TYPE_CODE_CONTEXT, 0, NULL);
	return self;
}


/**
 * Returns a copy of the list of source files.
 *
 * @return list of source files
 */
GeeCollection* vala_code_context_get_source_files (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), NULL);
	return GEE_COLLECTION (gee_read_only_collection_new (VALA_TYPE_SOURCE_FILE, ((GBoxedCopyFunc) g_object_ref), g_object_unref, GEE_COLLECTION (self->priv->source_files)));
}


/**
 * Returns a copy of the list of C source files.
 *
 * @return list of C source files
 */
GeeCollection* vala_code_context_get_c_source_files (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), NULL);
	return GEE_COLLECTION (gee_read_only_collection_new (G_TYPE_STRING, ((GBoxedCopyFunc) g_strdup), g_free, GEE_COLLECTION (self->priv->c_source_files)));
}


/**
 * Adds the specified file to the list of source files.
 *
 * @param file a source file
 */
void vala_code_context_add_source_file (ValaCodeContext* self, ValaSourceFile* file) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	g_return_if_fail (VALA_IS_SOURCE_FILE (file));
	gee_collection_add (GEE_COLLECTION (self->priv->source_files), file);
}


/**
 * Adds the specified file to the list of C source files.
 *
 * @param file a C source file
 */
void vala_code_context_add_c_source_file (ValaCodeContext* self, const char* file) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	g_return_if_fail (file != NULL);
	gee_collection_add (GEE_COLLECTION (self->priv->c_source_files), file);
}


/**
 * Returns a copy of the list of used packages.
 *
 * @return list of used packages
 */
GeeCollection* vala_code_context_get_packages (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), NULL);
	return GEE_COLLECTION (gee_read_only_collection_new (G_TYPE_STRING, ((GBoxedCopyFunc) g_strdup), g_free, GEE_COLLECTION (self->priv->packages)));
}


/**
 * Returns whether the specified package is being used.
 *
 * @param pkg a package name
 * @return    true if the specified package is being used
 */
gboolean vala_code_context_has_package (ValaCodeContext* self, const char* pkg) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	g_return_val_if_fail (pkg != NULL, FALSE);
	return gee_collection_contains (GEE_COLLECTION (self->priv->packages), pkg);
}


/**
 * Adds the specified package to the list of used packages.
 *
 * @param pkg a package name
 */
void vala_code_context_add_package (ValaCodeContext* self, const char* pkg) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	g_return_if_fail (pkg != NULL);
	gee_collection_add (GEE_COLLECTION (self->priv->packages), pkg);
}


/**
 * Visits the complete code tree file by file.
 *
 * @param visitor the visitor to be called when traversing
 */
void vala_code_context_accept (ValaCodeContext* self, ValaCodeVisitor* visitor) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	g_return_if_fail (VALA_IS_CODE_VISITOR (visitor));
	vala_code_node_accept (VALA_CODE_NODE (vala_code_context_get_root (self)), visitor);
	{
		GeeList* file_collection;
		GeeIterator* file_it;
		file_collection = self->priv->source_files;
		file_it = gee_iterable_iterator (GEE_ITERABLE (file_collection));
		while (gee_iterator_next (file_it)) {
			ValaSourceFile* file;
			file = ((ValaSourceFile*) gee_iterator_get (file_it));
			{
				vala_source_file_accept (file, visitor);
				(file == NULL ? NULL : (file = (g_object_unref (file), NULL)));
			}
		}
		(file_it == NULL ? NULL : (file_it = (g_object_unref (file_it), NULL)));
	}
}


/**
 * Find and resolve cycles in source file dependencies.
 */
void vala_code_context_find_header_cycles (ValaCodeContext* self) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	/* find cycles in dependencies between source files */
	{
		GeeList* file_collection;
		GeeIterator* file_it;
		file_collection = self->priv->source_files;
		file_it = gee_iterable_iterator (GEE_ITERABLE (file_collection));
		while (gee_iterator_next (file_it)) {
			ValaSourceFile* file;
			file = ((ValaSourceFile*) gee_iterator_get (file_it));
			{
				/* we're only interested in internal source files */
				if (vala_source_file_get_external_package (file)) {
					(file == NULL ? NULL : (file = (g_object_unref (file), NULL)));
					continue;
				}
				if (vala_source_file_get_mark (file) == 0) {
					GeeArrayList* _tmp0;
					_tmp0 = NULL;
					vala_code_context_visit (self, file, GEE_COLLECTION ((_tmp0 = gee_array_list_new (VALA_TYPE_SOURCE_FILE, ((GBoxedCopyFunc) g_object_ref), g_object_unref, g_direct_equal))));
					(_tmp0 == NULL ? NULL : (_tmp0 = (g_object_unref (_tmp0), NULL)));
				}
				(file == NULL ? NULL : (file = (g_object_unref (file), NULL)));
			}
		}
		(file_it == NULL ? NULL : (file_it = (g_object_unref (file_it), NULL)));
	}
	/* find one head for each cycle, it must not have any
	 * hard dependencies on other files in the cycle
	 */
	{
		GeeList* cycle_collection;
		GeeIterator* cycle_it;
		cycle_collection = self->priv->cycles;
		cycle_it = gee_iterable_iterator (GEE_ITERABLE (cycle_collection));
		while (gee_iterator_next (cycle_it)) {
			ValaSourceFileCycle* cycle;
			cycle = ((ValaSourceFileCycle*) gee_iterator_get (cycle_it));
			{
				cycle->head = vala_code_context_find_cycle_head (self, ((ValaSourceFile*) gee_list_get (cycle->files, 0)));
				vala_source_file_set_is_cycle_head (cycle->head, TRUE);
				(cycle == NULL ? NULL : (cycle = (g_object_unref (cycle), NULL)));
			}
		}
		(cycle_it == NULL ? NULL : (cycle_it = (g_object_unref (cycle_it), NULL)));
	}
	/* connect the source files in a non-cyclic way
	 * cycle members connect to the head of their cycle
	 */
	{
		GeeList* file2_collection;
		GeeIterator* file2_it;
		file2_collection = self->priv->source_files;
		file2_it = gee_iterable_iterator (GEE_ITERABLE (file2_collection));
		while (gee_iterator_next (file2_it)) {
			ValaSourceFile* file2;
			file2 = ((ValaSourceFile*) gee_iterator_get (file2_it));
			{
				/* we're only interested in internal source files */
				if (vala_source_file_get_external_package (file2)) {
					(file2 == NULL ? NULL : (file2 = (g_object_unref (file2), NULL)));
					continue;
				}
				{
					GeeCollection* dep_collection;
					GeeIterator* dep_it;
					dep_collection = vala_source_file_get_header_internal_dependencies (file2);
					dep_it = gee_iterable_iterator (GEE_ITERABLE (dep_collection));
					while (gee_iterator_next (dep_it)) {
						ValaSourceFile* _tmp3;
						ValaSourceFile* dep;
						_tmp3 = NULL;
						dep = (_tmp3 = ((ValaSourceFile*) gee_iterator_get (dep_it)), (_tmp3 == NULL ? NULL : g_object_ref (_tmp3)));
						{
							if (vala_source_file_get_cycle (file2) != NULL && vala_source_file_get_cycle (dep) == vala_source_file_get_cycle (file2)) {
								/* in the same cycle */
								if (!vala_source_file_get_is_cycle_head (file2)) {
									char* _tmp1;
									/* include header of cycle head */
									_tmp1 = NULL;
									vala_source_file_add_header_internal_include (file2, (_tmp1 = vala_source_file_get_cinclude_filename (vala_source_file_get_cycle (file2)->head)));
									_tmp1 = (g_free (_tmp1), NULL);
								}
							} else {
								char* _tmp2;
								/* we can just include the headers if they are not in a cycle or not in the same cycle as the current file */
								_tmp2 = NULL;
								vala_source_file_add_header_internal_include (file2, (_tmp2 = vala_source_file_get_cinclude_filename (dep)));
								_tmp2 = (g_free (_tmp2), NULL);
							}
							(dep == NULL ? NULL : (dep = (g_object_unref (dep), NULL)));
						}
					}
					(dep_collection == NULL ? NULL : (dep_collection = (g_object_unref (dep_collection), NULL)));
					(dep_it == NULL ? NULL : (dep_it = (g_object_unref (dep_it), NULL)));
				}
				(file2 == NULL ? NULL : (file2 = (g_object_unref (file2), NULL)));
			}
		}
		(file2_it == NULL ? NULL : (file2_it = (g_object_unref (file2_it), NULL)));
	}
}


static ValaSourceFile* vala_code_context_find_cycle_head (ValaCodeContext* self, ValaSourceFile* file) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), NULL);
	g_return_val_if_fail (VALA_IS_SOURCE_FILE (file), NULL);
	{
		GeeCollection* dep_collection;
		GeeIterator* dep_it;
		dep_collection = vala_source_file_get_header_internal_full_dependencies (file);
		dep_it = gee_iterable_iterator (GEE_ITERABLE (dep_collection));
		while (gee_iterator_next (dep_it)) {
			ValaSourceFile* _tmp2;
			ValaSourceFile* dep;
			_tmp2 = NULL;
			dep = (_tmp2 = ((ValaSourceFile*) gee_iterator_get (dep_it)), (_tmp2 == NULL ? NULL : g_object_ref (_tmp2)));
			{
				if (dep == file) {
					/* ignore file-internal dependencies */
					(dep == NULL ? NULL : (dep = (g_object_unref (dep), NULL)));
					continue;
				}
				{
					GeeList* cycle_file_collection;
					GeeIterator* cycle_file_it;
					cycle_file_collection = vala_source_file_get_cycle (file)->files;
					cycle_file_it = gee_iterable_iterator (GEE_ITERABLE (cycle_file_collection));
					while (gee_iterator_next (cycle_file_it)) {
						ValaSourceFile* _tmp1;
						ValaSourceFile* cycle_file;
						_tmp1 = NULL;
						cycle_file = (_tmp1 = ((ValaSourceFile*) gee_iterator_get (cycle_file_it)), (_tmp1 == NULL ? NULL : g_object_ref (_tmp1)));
						{
							if (dep == cycle_file) {
								ValaSourceFile* _tmp0;
								_tmp0 = NULL;
								return (_tmp0 = vala_code_context_find_cycle_head (self, dep), (cycle_file == NULL ? NULL : (cycle_file = (g_object_unref (cycle_file), NULL))), (cycle_file_it == NULL ? NULL : (cycle_file_it = (g_object_unref (cycle_file_it), NULL))), (dep == NULL ? NULL : (dep = (g_object_unref (dep), NULL))), (dep_collection == NULL ? NULL : (dep_collection = (g_object_unref (dep_collection), NULL))), (dep_it == NULL ? NULL : (dep_it = (g_object_unref (dep_it), NULL))), _tmp0);
							}
							(cycle_file == NULL ? NULL : (cycle_file = (g_object_unref (cycle_file), NULL)));
						}
					}
					(cycle_file_it == NULL ? NULL : (cycle_file_it = (g_object_unref (cycle_file_it), NULL)));
				}
				(dep == NULL ? NULL : (dep = (g_object_unref (dep), NULL)));
			}
		}
		(dep_collection == NULL ? NULL : (dep_collection = (g_object_unref (dep_collection), NULL)));
		(dep_it == NULL ? NULL : (dep_it = (g_object_unref (dep_it), NULL)));
	}
	/* no hard dependencies on members of the same cycle found
	 * source file suitable as cycle head
	 */
	return file;
}


static void vala_code_context_visit (ValaCodeContext* self, ValaSourceFile* file, GeeCollection* chain) {
	GeeList* l;
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	g_return_if_fail (VALA_IS_SOURCE_FILE (file));
	g_return_if_fail (GEE_IS_COLLECTION (chain));
	l = GEE_LIST (gee_array_list_new (VALA_TYPE_SOURCE_FILE, ((GBoxedCopyFunc) g_object_ref), g_object_unref, g_direct_equal));
	{
		GeeCollection* chain_file_collection;
		GeeIterator* chain_file_it;
		chain_file_collection = chain;
		chain_file_it = gee_iterable_iterator (GEE_ITERABLE (chain_file_collection));
		while (gee_iterator_next (chain_file_it)) {
			ValaSourceFile* chain_file;
			chain_file = ((ValaSourceFile*) gee_iterator_get (chain_file_it));
			{
				gee_collection_add (GEE_COLLECTION (l), chain_file);
				(chain_file == NULL ? NULL : (chain_file = (g_object_unref (chain_file), NULL)));
			}
		}
		(chain_file_it == NULL ? NULL : (chain_file_it = (g_object_unref (chain_file_it), NULL)));
	}
	gee_collection_add (GEE_COLLECTION (l), file);
	/* mark file as currently being visited */
	vala_source_file_set_mark (file, 1);
	{
		GeeCollection* dep_collection;
		GeeIterator* dep_it;
		dep_collection = vala_source_file_get_header_internal_dependencies (file);
		dep_it = gee_iterable_iterator (GEE_ITERABLE (dep_collection));
		while (gee_iterator_next (dep_it)) {
			ValaSourceFile* _tmp2;
			ValaSourceFile* dep;
			_tmp2 = NULL;
			dep = (_tmp2 = ((ValaSourceFile*) gee_iterator_get (dep_it)), (_tmp2 == NULL ? NULL : g_object_ref (_tmp2)));
			{
				if (file == dep) {
					(dep == NULL ? NULL : (dep = (g_object_unref (dep), NULL)));
					continue;
				}
				if (vala_source_file_get_mark (dep) == 1) {
					ValaSourceFileCycle* cycle;
					gboolean cycle_start_found;
					/* found cycle */
					cycle = vala_source_file_cycle_new ();
					gee_collection_add (GEE_COLLECTION (self->priv->cycles), cycle);
					cycle_start_found = FALSE;
					{
						GeeList* cycle_file_collection;
						GeeIterator* cycle_file_it;
						cycle_file_collection = l;
						cycle_file_it = gee_iterable_iterator (GEE_ITERABLE (cycle_file_collection));
						while (gee_iterator_next (cycle_file_it)) {
							ValaSourceFile* cycle_file;
							cycle_file = ((ValaSourceFile*) gee_iterator_get (cycle_file_it));
							{
								ValaSourceFileCycle* _tmp0;
								ValaSourceFileCycle* ref_cycle_file_cycle;
								_tmp0 = NULL;
								ref_cycle_file_cycle = (_tmp0 = vala_source_file_get_cycle (cycle_file), (_tmp0 == NULL ? NULL : g_object_ref (_tmp0)));
								if (!cycle_start_found) {
									if (cycle_file == dep) {
										cycle_start_found = TRUE;
									}
								}
								if (!cycle_start_found) {
									(cycle_file == NULL ? NULL : (cycle_file = (g_object_unref (cycle_file), NULL)));
									(ref_cycle_file_cycle == NULL ? NULL : (ref_cycle_file_cycle = (g_object_unref (ref_cycle_file_cycle), NULL)));
									continue;
								}
								if (vala_source_file_get_cycle (cycle_file) != NULL) {
									/* file already in a cycle */
									if (vala_source_file_get_cycle (cycle_file) == cycle) {
										/* file is in the same cycle, nothing to do */
										(cycle_file == NULL ? NULL : (cycle_file = (g_object_unref (cycle_file), NULL)));
										(ref_cycle_file_cycle == NULL ? NULL : (ref_cycle_file_cycle = (g_object_unref (ref_cycle_file_cycle), NULL)));
										continue;
									}
									/* file is in an other cycle, merge the two cycles */
									gee_collection_remove (GEE_COLLECTION (self->priv->cycles), vala_source_file_get_cycle (cycle_file));
									{
										GeeList* inner_cycle_file_collection;
										GeeIterator* inner_cycle_file_it;
										inner_cycle_file_collection = vala_source_file_get_cycle (cycle_file)->files;
										inner_cycle_file_it = gee_iterable_iterator (GEE_ITERABLE (inner_cycle_file_collection));
										while (gee_iterator_next (inner_cycle_file_it)) {
											ValaSourceFile* _tmp1;
											ValaSourceFile* inner_cycle_file;
											_tmp1 = NULL;
											inner_cycle_file = (_tmp1 = ((ValaSourceFile*) gee_iterator_get (inner_cycle_file_it)), (_tmp1 == NULL ? NULL : g_object_ref (_tmp1)));
											{
												if (vala_source_file_get_cycle (inner_cycle_file) != cycle) {
													/* file in inner cycle not yet added to outer cycle */
													gee_collection_add (GEE_COLLECTION (cycle->files), inner_cycle_file);
													vala_source_file_set_cycle (inner_cycle_file, cycle);
												}
												(inner_cycle_file == NULL ? NULL : (inner_cycle_file = (g_object_unref (inner_cycle_file), NULL)));
											}
										}
										(inner_cycle_file_it == NULL ? NULL : (inner_cycle_file_it = (g_object_unref (inner_cycle_file_it), NULL)));
									}
								} else {
									gee_collection_add (GEE_COLLECTION (cycle->files), cycle_file);
									vala_source_file_set_cycle (cycle_file, cycle);
								}
								(cycle_file == NULL ? NULL : (cycle_file = (g_object_unref (cycle_file), NULL)));
								(ref_cycle_file_cycle == NULL ? NULL : (ref_cycle_file_cycle = (g_object_unref (ref_cycle_file_cycle), NULL)));
							}
						}
						(cycle_file_it == NULL ? NULL : (cycle_file_it = (g_object_unref (cycle_file_it), NULL)));
					}
					(cycle == NULL ? NULL : (cycle = (g_object_unref (cycle), NULL)));
				} else {
					if (vala_source_file_get_mark (dep) == 0) {
						/* found not yet visited file */
						vala_code_context_visit (self, dep, GEE_COLLECTION (l));
					}
				}
				(dep == NULL ? NULL : (dep = (g_object_unref (dep), NULL)));
			}
		}
		(dep_collection == NULL ? NULL : (dep_collection = (g_object_unref (dep_collection), NULL)));
		(dep_it == NULL ? NULL : (dep_it = (g_object_unref (dep_it), NULL)));
	}
	/* mark file as successfully visited */
	vala_source_file_set_mark (file, 2);
	(l == NULL ? NULL : (l = (g_object_unref (l), NULL)));
}


void vala_code_context_add_define (ValaCodeContext* self, const char* define) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	g_return_if_fail (define != NULL);
	gee_collection_add (GEE_COLLECTION (self->priv->defines), define);
}


gboolean vala_code_context_ignore_node (ValaCodeContext* self, ValaCodeNode* node) {
	ValaAttribute* attr;
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	g_return_val_if_fail (VALA_IS_CODE_NODE (node), FALSE);
	attr = vala_code_node_get_attribute (node, "Conditional");
	if (attr == NULL) {
		gboolean _tmp0;
		return (_tmp0 = FALSE, (attr == NULL ? NULL : (attr = (g_object_unref (attr), NULL))), _tmp0);
	} else {
		char* condition;
		condition = vala_attribute_get_string (attr, "condition");
		if (condition == NULL) {
			gboolean _tmp1;
			return (_tmp1 = FALSE, (condition = (g_free (condition), NULL)), (attr == NULL ? NULL : (attr = (g_object_unref (attr), NULL))), _tmp1);
		} else {
			gboolean _tmp2;
			return (_tmp2 = !gee_collection_contains (GEE_COLLECTION (self->priv->defines), condition), (condition = (g_free (condition), NULL)), (attr == NULL ? NULL : (attr = (g_object_unref (attr), NULL))), _tmp2);
		}
		condition = (g_free (condition), NULL);
	}
	(attr == NULL ? NULL : (attr = (g_object_unref (attr), NULL)));
}


char* vala_code_context_get_package_path (ValaCodeContext* self, const char* pkg, int vapi_directories_length1, char** vapi_directories) {
	char* basename;
	char* filename;
	char* _tmp3;
	gpointer _tmp8;
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), NULL);
	g_return_val_if_fail (pkg != NULL, NULL);
	basename = g_strdup_printf ("%s.vapi", pkg);
	filename = NULL;
	if (vapi_directories != NULL) {
		{
			char** vapidir_collection;
			int vapidir_it;
			vapidir_collection = vapi_directories;
			for (vapidir_it = 0; (vapi_directories_length1 != -1 && vapidir_it < vapi_directories_length1) || (vapi_directories_length1 == -1 && vapidir_collection[vapidir_it] != NULL); vapidir_it = vapidir_it + 1) {
				const char* _tmp2;
				char* vapidir;
				_tmp2 = NULL;
				vapidir = (_tmp2 = vapidir_collection[vapidir_it], (_tmp2 == NULL ? NULL : g_strdup (_tmp2)));
				{
					char* _tmp0;
					_tmp0 = NULL;
					filename = (_tmp0 = g_build_filename (vapidir, basename, NULL), (filename = (g_free (filename), NULL)), _tmp0);
					if (g_file_test (filename, G_FILE_TEST_EXISTS)) {
						char* _tmp1;
						_tmp1 = NULL;
						return (_tmp1 = filename, (vapidir = (g_free (vapidir), NULL)), (basename = (g_free (basename), NULL)), _tmp1);
					}
					vapidir = (g_free (vapidir), NULL);
				}
			}
		}
	}
	_tmp3 = NULL;
	filename = (_tmp3 = g_build_filename (PACKAGE_DATADIR, "vapi", basename, NULL), (filename = (g_free (filename), NULL)), _tmp3);
	if (g_file_test (filename, G_FILE_TEST_EXISTS)) {
		char* _tmp4;
		_tmp4 = NULL;
		return (_tmp4 = filename, (basename = (g_free (basename), NULL)), _tmp4);
	}
	{
		char** vapidir_collection;
		char** vapidir_it;
		int vapidir_collection_length1;
		vapidir_collection = g_get_system_data_dirs ();
		vapidir_collection_length1 = -1;
		for (vapidir_it = vapidir_collection; *vapidir_it != NULL; vapidir_it = vapidir_it + 1) {
			const char* _tmp7;
			char* vapidir;
			_tmp7 = NULL;
			vapidir = (_tmp7 = *vapidir_it, (_tmp7 == NULL ? NULL : g_strdup (_tmp7)));
			{
				char* _tmp5;
				_tmp5 = NULL;
				filename = (_tmp5 = g_build_filename (vapidir, "vala/vapi", basename, NULL), (filename = (g_free (filename), NULL)), _tmp5);
				if (g_file_test (filename, G_FILE_TEST_EXISTS)) {
					char* _tmp6;
					_tmp6 = NULL;
					return (_tmp6 = filename, (vapidir = (g_free (vapidir), NULL)), (basename = (g_free (basename), NULL)), _tmp6);
				}
				vapidir = (g_free (vapidir), NULL);
			}
		}
	}
	return (_tmp8 = NULL, (basename = (g_free (basename), NULL)), (filename = (g_free (filename), NULL)), _tmp8);
}


void vala_code_context_add_dynamic_member (ValaCodeContext* self, ValaSymbol* dynamic_member) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	g_return_if_fail (VALA_IS_SYMBOL (dynamic_member));
	gee_collection_add (GEE_COLLECTION (self->priv->dynamic_members), dynamic_member);
}


char* vala_code_context_get_library (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), NULL);
	return self->priv->_library;
}


void vala_code_context_set_library (ValaCodeContext* self, const char* value) {
	char* _tmp2;
	const char* _tmp1;
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_library = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL ? NULL : g_strdup (_tmp1))), (self->priv->_library = (g_free (self->priv->_library), NULL)), _tmp2);
}


gboolean vala_code_context_get_memory_management (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	return self->priv->_memory_management;
}


void vala_code_context_set_memory_management (ValaCodeContext* self, gboolean value) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	self->priv->_memory_management = value;
}


gboolean vala_code_context_get_assert (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	return self->priv->_assert;
}


void vala_code_context_set_assert (ValaCodeContext* self, gboolean value) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	self->priv->_assert = value;
}


gboolean vala_code_context_get_checking (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	return self->priv->_checking;
}


void vala_code_context_set_checking (ValaCodeContext* self, gboolean value) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	self->priv->_checking = value;
}


gboolean vala_code_context_get_non_null (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	return self->priv->_non_null;
}


void vala_code_context_set_non_null (ValaCodeContext* self, gboolean value) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	self->priv->_non_null = value;
}


gboolean vala_code_context_get_non_null_experimental (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	return self->priv->_non_null_experimental;
}


void vala_code_context_set_non_null_experimental (ValaCodeContext* self, gboolean value) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	self->priv->_non_null_experimental = value;
}


gboolean vala_code_context_get_ccode_only (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	return self->priv->_ccode_only;
}


void vala_code_context_set_ccode_only (ValaCodeContext* self, gboolean value) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	self->priv->_ccode_only = value;
}


gboolean vala_code_context_get_compile_only (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	return self->priv->_compile_only;
}


void vala_code_context_set_compile_only (ValaCodeContext* self, gboolean value) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	self->priv->_compile_only = value;
}


char* vala_code_context_get_output (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), NULL);
	return self->priv->_output;
}


void vala_code_context_set_output (ValaCodeContext* self, const char* value) {
	char* _tmp2;
	const char* _tmp1;
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_output = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL ? NULL : g_strdup (_tmp1))), (self->priv->_output = (g_free (self->priv->_output), NULL)), _tmp2);
}


char* vala_code_context_get_basedir (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), NULL);
	return self->priv->_basedir;
}


void vala_code_context_set_basedir (ValaCodeContext* self, const char* value) {
	char* _tmp2;
	const char* _tmp1;
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_basedir = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL ? NULL : g_strdup (_tmp1))), (self->priv->_basedir = (g_free (self->priv->_basedir), NULL)), _tmp2);
}


char* vala_code_context_get_directory (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), NULL);
	return self->priv->_directory;
}


void vala_code_context_set_directory (ValaCodeContext* self, const char* value) {
	char* _tmp2;
	const char* _tmp1;
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_directory = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL ? NULL : g_strdup (_tmp1))), (self->priv->_directory = (g_free (self->priv->_directory), NULL)), _tmp2);
}


gboolean vala_code_context_get_debug (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	return self->priv->_debug;
}


void vala_code_context_set_debug (ValaCodeContext* self, gboolean value) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	self->priv->_debug = value;
}


gint vala_code_context_get_optlevel (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), 0);
	return self->priv->_optlevel;
}


void vala_code_context_set_optlevel (ValaCodeContext* self, gint value) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	self->priv->_optlevel = value;
}


gboolean vala_code_context_get_thread (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	return self->priv->_thread;
}


void vala_code_context_set_thread (ValaCodeContext* self, gboolean value) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	self->priv->_thread = value;
}


ValaMethod* vala_code_context_get_module_init_method (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), NULL);
	return self->priv->_module_init_method;
}


void vala_code_context_set_module_init_method (ValaCodeContext* self, ValaMethod* value) {
	ValaMethod* _tmp2;
	ValaMethod* _tmp1;
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_module_init_method = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL ? NULL : g_object_ref (_tmp1))), (self->priv->_module_init_method == NULL ? NULL : (self->priv->_module_init_method = (g_object_unref (self->priv->_module_init_method), NULL))), _tmp2);
}


gboolean vala_code_context_get_save_temps (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	return self->priv->_save_temps;
}


void vala_code_context_set_save_temps (ValaCodeContext* self, gboolean value) {
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	self->priv->_save_temps = value;
}


gboolean vala_code_context_get_save_csources (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	return self->priv->_save_temps;
}


gboolean vala_code_context_get_save_cheaders (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), FALSE);
	return vala_code_context_get_save_csources (self) || NULL != self->priv->_library;
}


ValaNamespace* vala_code_context_get_root (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), NULL);
	return self->priv->_root;
}


ValaCodeGenerator* vala_code_context_get_codegen (ValaCodeContext* self) {
	g_return_val_if_fail (VALA_IS_CODE_CONTEXT (self), NULL);
	return self->priv->_codegen;
}


void vala_code_context_set_codegen (ValaCodeContext* self, ValaCodeGenerator* value) {
	ValaCodeGenerator* _tmp2;
	ValaCodeGenerator* _tmp1;
	g_return_if_fail (VALA_IS_CODE_CONTEXT (self));
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_codegen = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL ? NULL : g_object_ref (_tmp1))), (self->priv->_codegen == NULL ? NULL : (self->priv->_codegen = (g_object_unref (self->priv->_codegen), NULL))), _tmp2);
}


static GObject * vala_code_context_constructor (GType type, guint n_construct_properties, GObjectConstructParam * construct_properties) {
	GObject * obj;
	ValaCodeContextClass * klass;
	GObjectClass * parent_class;
	ValaCodeContext * self;
	klass = VALA_CODE_CONTEXT_CLASS (g_type_class_peek (VALA_TYPE_CODE_CONTEXT));
	parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
	obj = parent_class->constructor (type, n_construct_properties, construct_properties);
	self = VALA_CODE_CONTEXT (obj);
	{
		ValaCodeGenerator* _tmp0;
		_tmp0 = NULL;
		vala_code_context_set_codegen (self, (_tmp0 = vala_code_generator_new ()));
		(_tmp0 == NULL ? NULL : (_tmp0 = (g_object_unref (_tmp0), NULL)));
	}
	return obj;
}


static void vala_code_context_get_property (GObject * object, guint property_id, GValue * value, GParamSpec * pspec) {
	ValaCodeContext * self;
	self = VALA_CODE_CONTEXT (object);
	switch (property_id) {
		case VALA_CODE_CONTEXT_LIBRARY:
		g_value_set_string (value, vala_code_context_get_library (self));
		break;
		case VALA_CODE_CONTEXT_MEMORY_MANAGEMENT:
		g_value_set_boolean (value, vala_code_context_get_memory_management (self));
		break;
		case VALA_CODE_CONTEXT_ASSERT:
		g_value_set_boolean (value, vala_code_context_get_assert (self));
		break;
		case VALA_CODE_CONTEXT_CHECKING:
		g_value_set_boolean (value, vala_code_context_get_checking (self));
		break;
		case VALA_CODE_CONTEXT_NON_NULL:
		g_value_set_boolean (value, vala_code_context_get_non_null (self));
		break;
		case VALA_CODE_CONTEXT_NON_NULL_EXPERIMENTAL:
		g_value_set_boolean (value, vala_code_context_get_non_null_experimental (self));
		break;
		case VALA_CODE_CONTEXT_CCODE_ONLY:
		g_value_set_boolean (value, vala_code_context_get_ccode_only (self));
		break;
		case VALA_CODE_CONTEXT_COMPILE_ONLY:
		g_value_set_boolean (value, vala_code_context_get_compile_only (self));
		break;
		case VALA_CODE_CONTEXT_OUTPUT:
		g_value_set_string (value, vala_code_context_get_output (self));
		break;
		case VALA_CODE_CONTEXT_BASEDIR:
		g_value_set_string (value, vala_code_context_get_basedir (self));
		break;
		case VALA_CODE_CONTEXT_DIRECTORY:
		g_value_set_string (value, vala_code_context_get_directory (self));
		break;
		case VALA_CODE_CONTEXT_DEBUG:
		g_value_set_boolean (value, vala_code_context_get_debug (self));
		break;
		case VALA_CODE_CONTEXT_OPTLEVEL:
		g_value_set_int (value, vala_code_context_get_optlevel (self));
		break;
		case VALA_CODE_CONTEXT_THREAD:
		g_value_set_boolean (value, vala_code_context_get_thread (self));
		break;
		case VALA_CODE_CONTEXT_MODULE_INIT_METHOD:
		g_value_set_object (value, vala_code_context_get_module_init_method (self));
		break;
		case VALA_CODE_CONTEXT_SAVE_TEMPS:
		g_value_set_boolean (value, vala_code_context_get_save_temps (self));
		break;
		case VALA_CODE_CONTEXT_SAVE_CSOURCES:
		g_value_set_boolean (value, vala_code_context_get_save_csources (self));
		break;
		case VALA_CODE_CONTEXT_SAVE_CHEADERS:
		g_value_set_boolean (value, vala_code_context_get_save_cheaders (self));
		break;
		case VALA_CODE_CONTEXT_ROOT:
		g_value_set_object (value, vala_code_context_get_root (self));
		break;
		case VALA_CODE_CONTEXT_CODEGEN:
		g_value_set_object (value, vala_code_context_get_codegen (self));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void vala_code_context_set_property (GObject * object, guint property_id, const GValue * value, GParamSpec * pspec) {
	ValaCodeContext * self;
	self = VALA_CODE_CONTEXT (object);
	switch (property_id) {
		case VALA_CODE_CONTEXT_LIBRARY:
		vala_code_context_set_library (self, g_value_get_string (value));
		break;
		case VALA_CODE_CONTEXT_MEMORY_MANAGEMENT:
		vala_code_context_set_memory_management (self, g_value_get_boolean (value));
		break;
		case VALA_CODE_CONTEXT_ASSERT:
		vala_code_context_set_assert (self, g_value_get_boolean (value));
		break;
		case VALA_CODE_CONTEXT_CHECKING:
		vala_code_context_set_checking (self, g_value_get_boolean (value));
		break;
		case VALA_CODE_CONTEXT_NON_NULL:
		vala_code_context_set_non_null (self, g_value_get_boolean (value));
		break;
		case VALA_CODE_CONTEXT_NON_NULL_EXPERIMENTAL:
		vala_code_context_set_non_null_experimental (self, g_value_get_boolean (value));
		break;
		case VALA_CODE_CONTEXT_CCODE_ONLY:
		vala_code_context_set_ccode_only (self, g_value_get_boolean (value));
		break;
		case VALA_CODE_CONTEXT_COMPILE_ONLY:
		vala_code_context_set_compile_only (self, g_value_get_boolean (value));
		break;
		case VALA_CODE_CONTEXT_OUTPUT:
		vala_code_context_set_output (self, g_value_get_string (value));
		break;
		case VALA_CODE_CONTEXT_BASEDIR:
		vala_code_context_set_basedir (self, g_value_get_string (value));
		break;
		case VALA_CODE_CONTEXT_DIRECTORY:
		vala_code_context_set_directory (self, g_value_get_string (value));
		break;
		case VALA_CODE_CONTEXT_DEBUG:
		vala_code_context_set_debug (self, g_value_get_boolean (value));
		break;
		case VALA_CODE_CONTEXT_OPTLEVEL:
		vala_code_context_set_optlevel (self, g_value_get_int (value));
		break;
		case VALA_CODE_CONTEXT_THREAD:
		vala_code_context_set_thread (self, g_value_get_boolean (value));
		break;
		case VALA_CODE_CONTEXT_MODULE_INIT_METHOD:
		vala_code_context_set_module_init_method (self, g_value_get_object (value));
		break;
		case VALA_CODE_CONTEXT_SAVE_TEMPS:
		vala_code_context_set_save_temps (self, g_value_get_boolean (value));
		break;
		case VALA_CODE_CONTEXT_CODEGEN:
		vala_code_context_set_codegen (self, g_value_get_object (value));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void vala_code_context_class_init (ValaCodeContextClass * klass) {
	vala_code_context_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (ValaCodeContextPrivate));
	G_OBJECT_CLASS (klass)->get_property = vala_code_context_get_property;
	G_OBJECT_CLASS (klass)->set_property = vala_code_context_set_property;
	G_OBJECT_CLASS (klass)->constructor = vala_code_context_constructor;
	G_OBJECT_CLASS (klass)->dispose = vala_code_context_dispose;
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_LIBRARY, g_param_spec_string ("library", "library", "library", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_MEMORY_MANAGEMENT, g_param_spec_boolean ("memory-management", "memory-management", "memory-management", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_ASSERT, g_param_spec_boolean ("assert", "assert", "assert", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_CHECKING, g_param_spec_boolean ("checking", "checking", "checking", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_NON_NULL, g_param_spec_boolean ("non-null", "non-null", "non-null", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_NON_NULL_EXPERIMENTAL, g_param_spec_boolean ("non-null-experimental", "non-null-experimental", "non-null-experimental", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_CCODE_ONLY, g_param_spec_boolean ("ccode-only", "ccode-only", "ccode-only", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_COMPILE_ONLY, g_param_spec_boolean ("compile-only", "compile-only", "compile-only", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_OUTPUT, g_param_spec_string ("output", "output", "output", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_BASEDIR, g_param_spec_string ("basedir", "basedir", "basedir", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_DIRECTORY, g_param_spec_string ("directory", "directory", "directory", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_DEBUG, g_param_spec_boolean ("debug", "debug", "debug", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_OPTLEVEL, g_param_spec_int ("optlevel", "optlevel", "optlevel", G_MININT, G_MAXINT, 0, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_THREAD, g_param_spec_boolean ("thread", "thread", "thread", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_MODULE_INIT_METHOD, g_param_spec_object ("module-init-method", "module-init-method", "module-init-method", VALA_TYPE_METHOD, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_SAVE_TEMPS, g_param_spec_boolean ("save-temps", "save-temps", "save-temps", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_SAVE_CSOURCES, g_param_spec_boolean ("save-csources", "save-csources", "save-csources", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_SAVE_CHEADERS, g_param_spec_boolean ("save-cheaders", "save-cheaders", "save-cheaders", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_ROOT, g_param_spec_object ("root", "root", "root", VALA_TYPE_NAMESPACE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_CODE_CONTEXT_CODEGEN, g_param_spec_object ("codegen", "codegen", "codegen", VALA_TYPE_CODE_GENERATOR, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
}


static void vala_code_context_init (ValaCodeContext * self) {
	self->priv = VALA_CODE_CONTEXT_GET_PRIVATE (self);
	self->priv->source_files = GEE_LIST (gee_array_list_new (VALA_TYPE_SOURCE_FILE, ((GBoxedCopyFunc) g_object_ref), g_object_unref, g_direct_equal));
	self->priv->c_source_files = GEE_LIST (gee_array_list_new (G_TYPE_STRING, ((GBoxedCopyFunc) g_strdup), g_free, g_direct_equal));
	self->priv->_root = vala_namespace_new (NULL, NULL);
	self->priv->cycles = GEE_LIST (gee_array_list_new (VALA_TYPE_SOURCE_FILE_CYCLE, ((GBoxedCopyFunc) g_object_ref), g_object_unref, g_direct_equal));
	self->priv->packages = GEE_LIST (gee_array_list_new (G_TYPE_STRING, ((GBoxedCopyFunc) g_strdup), g_free, g_str_equal));
	self->priv->defines = GEE_LIST (gee_array_list_new (G_TYPE_STRING, ((GBoxedCopyFunc) g_strdup), g_free, g_str_equal));
	self->priv->dynamic_members = GEE_LIST (gee_array_list_new (VALA_TYPE_SYMBOL, ((GBoxedCopyFunc) g_object_ref), g_object_unref, g_direct_equal));
}


static void vala_code_context_dispose (GObject * obj) {
	ValaCodeContext * self;
	self = VALA_CODE_CONTEXT (obj);
	self->priv->_library = (g_free (self->priv->_library), NULL);
	self->priv->_output = (g_free (self->priv->_output), NULL);
	self->priv->_basedir = (g_free (self->priv->_basedir), NULL);
	self->priv->_directory = (g_free (self->priv->_directory), NULL);
	(self->priv->_module_init_method == NULL ? NULL : (self->priv->_module_init_method = (g_object_unref (self->priv->_module_init_method), NULL)));
	(self->priv->source_files == NULL ? NULL : (self->priv->source_files = (g_object_unref (self->priv->source_files), NULL)));
	(self->priv->c_source_files == NULL ? NULL : (self->priv->c_source_files = (g_object_unref (self->priv->c_source_files), NULL)));
	(self->priv->_root == NULL ? NULL : (self->priv->_root = (g_object_unref (self->priv->_root), NULL)));
	(self->priv->cycles == NULL ? NULL : (self->priv->cycles = (g_object_unref (self->priv->cycles), NULL)));
	(self->priv->packages == NULL ? NULL : (self->priv->packages = (g_object_unref (self->priv->packages), NULL)));
	(self->priv->defines == NULL ? NULL : (self->priv->defines = (g_object_unref (self->priv->defines), NULL)));
	(self->priv->dynamic_members == NULL ? NULL : (self->priv->dynamic_members = (g_object_unref (self->priv->dynamic_members), NULL)));
	(self->priv->_codegen == NULL ? NULL : (self->priv->_codegen = (g_object_unref (self->priv->_codegen), NULL)));
	G_OBJECT_CLASS (vala_code_context_parent_class)->dispose (obj);
}


GType vala_code_context_get_type (void) {
	static GType vala_code_context_type_id = 0;
	if (G_UNLIKELY (vala_code_context_type_id == 0)) {
		static const GTypeInfo g_define_type_info = { sizeof (ValaCodeContextClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) vala_code_context_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (ValaCodeContext), 0, (GInstanceInitFunc) vala_code_context_init };
		vala_code_context_type_id = g_type_register_static (G_TYPE_OBJECT, "ValaCodeContext", &g_define_type_info, 0);
	}
	return vala_code_context_type_id;
}




